@@ -23,6 +23,7 @@ import urllib
 
 import cherrypy
 
+import large_image
 from girder.api import access, filter_logging
 from girder.api.describe import Description, autoDescribeRoute, describeRoute
 from girder.api.rest import filtermodel, loadmodel, setRawResponse, setResponseHeader
@@ -34,7 +35,7 @@ from girder.models.item import Item
 from girder.utility.progress import setResponseTimeLimit
 from large_image.cache_util import strhash
 from large_image.constants import TileInputUnits
-from large_image.exceptions import TileGeneralException
+from large_image.exceptions import TileGeneralError
 
 from .. import loadmodelcache
 from ..models.image_item import ImageItem
@@ -103,29 +104,25 @@ class TilesItemResource(ItemResource):
         apiRoot.item.route('POST', (':itemId', 'tiles', 'convert'), self.convertImage)
         apiRoot.item.route('GET', (':itemId', 'tiles'), self.getTilesInfo)
         apiRoot.item.route('DELETE', (':itemId', 'tiles'), self.deleteTiles)
-        apiRoot.item.route('GET', (':itemId', 'tiles', 'thumbnail'),
-                           self.getTilesThumbnail)
-        apiRoot.item.route('GET', (':itemId', 'tiles', 'region'),
-                           self.getTilesRegion)
-        apiRoot.item.route('GET', (':itemId', 'tiles', 'pixel'),
-                           self.getTilesPixel)
-        apiRoot.item.route('GET', (':itemId', 'tiles', 'histogram'),
-                           self.getHistogram)
-        apiRoot.item.route('GET', (':itemId', 'tiles', 'zxy', ':z', ':x', ':y'),
-                           self.getTile)
+        apiRoot.item.route('GET', (':itemId', 'tiles', 'thumbnail'), self.getTilesThumbnail)
+        apiRoot.item.route('GET', (':itemId', 'tiles', 'region'), self.getTilesRegion)
+        apiRoot.item.route('GET', (':itemId', 'tiles', 'tile_frames'), self.tileFrames)
+        apiRoot.item.route('GET', (':itemId', 'tiles', 'tile_frames', 'quad_info'),
+                           self.tileFramesQuadInfo)
+        apiRoot.item.route('GET', (':itemId', 'tiles', 'pixel'), self.getTilesPixel)
+        apiRoot.item.route('GET', (':itemId', 'tiles', 'histogram'), self.getHistogram)
+        apiRoot.item.route('GET', (':itemId', 'tiles', 'bands'), self.getBandInformation)
+        apiRoot.item.route('GET', (':itemId', 'tiles', 'zxy', ':z', ':x', ':y'), self.getTile)
         apiRoot.item.route('GET', (':itemId', 'tiles', 'fzxy', ':frame', ':z', ':x', ':y'),
                            self.getTileWithFrame)
-        apiRoot.item.route('GET', (':itemId', 'tiles', 'images'),
-                           self.getAssociatedImagesList)
+        apiRoot.item.route('GET', (':itemId', 'tiles', 'images'), self.getAssociatedImagesList)
         apiRoot.item.route('GET', (':itemId', 'tiles', 'images', ':image'),
                            self.getAssociatedImage)
         apiRoot.item.route('GET', (':itemId', 'tiles', 'images', ':image', 'metadata'),
                            self.getAssociatedImageMetadata)
         apiRoot.item.route('GET', ('test', 'tiles'), self.getTestTilesInfo)
-        apiRoot.item.route('GET', ('test', 'tiles', 'zxy', ':z', ':x', ':y'),
-                           self.getTestTile)
-        apiRoot.item.route('GET', (':itemId', 'tiles', 'dzi.dzi'),
-                           self.getDZIInfo)
+        apiRoot.item.route('GET', ('test', 'tiles', 'zxy', ':z', ':x', ':y'), self.getTestTile)
+        apiRoot.item.route('GET', (':itemId', 'tiles', 'dzi.dzi'), self.getDZIInfo)
         apiRoot.item.route('GET', (':itemId', 'tiles', 'dzi_files', ':level', ':xandy'),
                            self.getDZITile)
         apiRoot.item.route('GET', (':itemId', 'tiles', 'internal_metadata'),
@@ -194,7 +191,7 @@ class TilesItemResource(ItemResource):
                 createJob='always' if self.boolParam('force', params, default=False) else True,
                 notify=notify,
                 **params)
-        except TileGeneralException as e:
+        except TileGeneralError as e:
             raise RestException(e.args[0])
 
     @describeRoute(
@@ -257,7 +254,7 @@ class TilesItemResource(ItemResource):
         try:
             return self.imageItemModel.convertImage(
                 item, largeImageFile, user, token, localJob=localJob, **params)
-        except TileGeneralException as e:
+        except TileGeneralError as e:
             raise RestException(e.args[0])
 
     @classmethod
@@ -271,6 +268,9 @@ class TilesItemResource(ItemResource):
             ('sizeX', int),
             ('sizeY', int),
             ('fractal', lambda val: val == 'true'),
+            ('frame', int),
+            ('frames', str),
+            ('monochrome', lambda val: val == 'true'),
             ('encoding', str),
         ])
 
@@ -328,7 +328,7 @@ class TilesItemResource(ItemResource):
         """
         try:
             return self.imageItemModel.getMetadata(item, **imageArgs)
-        except TileGeneralException as e:
+        except TileGeneralError as e:
             raise RestException(e.args[0], code=400)
 
     def _setContentDisposition(self, item, contentDisposition, mime, subname, fullFilename=None):
@@ -386,7 +386,7 @@ class TilesItemResource(ItemResource):
     def getInternalMetadata(self, item, params):
         try:
             return self.imageItemModel.getInternalMetadata(item, **params)
-        except TileGeneralException as e:
+        except TileGeneralError as e:
             raise RestException(e.args[0], code=400)
 
     @describeRoute(
@@ -465,7 +465,7 @@ class TilesItemResource(ItemResource):
             try:
                 tileData, tileMime = self.imageItemModel.getTile(
                     item, x, y, z, mayRedirect=mayRedirect, **imageArgs)
-            except TileGeneralException as e:
+            except TileGeneralError as e:
                 raise RestException(e.args[0], code=404)
         setResponseHeader('Content-Type', tileMime)
         setRawResponse()
@@ -693,7 +693,7 @@ class TilesItemResource(ItemResource):
         _handleETag('getTilesThumbnail', item, params)
         try:
             result = self.imageItemModel.getThumbnail(item, **params)
-        except TileGeneralException as e:
+        except TileGeneralError as e:
             raise RestException(e.args[0])
         except ValueError as e:
             raise RestException('Value Error: %s' % e.args[0])
@@ -831,7 +831,7 @@ class TilesItemResource(ItemResource):
         try:
             regionData, regionMime = self.imageItemModel.getRegion(
                 item, **params)
-        except TileGeneralException as e:
+        except TileGeneralError as e:
             raise RestException(e.args[0])
         except ValueError as e:
             raise RestException('Value Error: %s' % e.args[0])
@@ -887,7 +887,7 @@ class TilesItemResource(ItemResource):
         ])
         try:
             pixel = self.imageItemModel.getPixel(item, **params)
-        except TileGeneralException as e:
+        except TileGeneralError as e:
             raise RestException(e.args[0])
         except ValueError as e:
             raise RestException('Value Error: %s' % e.args[0])
@@ -977,6 +977,26 @@ class TilesItemResource(ItemResource):
                     entry[key] = float(entry[key])
         return result
 
+    @describeRoute(
+        Description('Get band information for a large image item.')
+        .param('itemId', 'The ID of the item.', paramType='path')
+        .param('frame', 'For multiframe images, the 0-based frame number.  '
+               'This is ignored on non-multiframe images.', required=False,
+               dataType='int')
+        .errorResponse('ID was invalid.')
+        .errorResponse('Read access was denied for the item.', 403)
+    )
+    @access.public
+    @loadmodel(model='item', map={'itemId': 'item'}, level=AccessType.READ)
+    def getBandInformation(self, item, params):
+        _adjustParams(params)
+        params = self._parseParams(params, True, [
+            ('frame', int),
+        ])
+        _handleETag('getBandInformation', item, params)
+        result = self.imageItemModel.getBandInformation(item, **params)
+        return result
+
     @describeRoute(
         Description('Get a list of additional images associated with a large image.')
         .param('itemId', 'The ID of the item.', paramType='path')
@@ -988,7 +1008,7 @@ class TilesItemResource(ItemResource):
     def getAssociatedImagesList(self, item, params):
         try:
             return self.imageItemModel.getAssociatedImagesList(item)
-        except TileGeneralException as e:
+        except TileGeneralError as e:
             raise RestException(e.args[0], code=400)
 
     @describeRoute(
@@ -1032,7 +1052,7 @@ class TilesItemResource(ItemResource):
         _handleETag('getAssociatedImage', item, image, params)
         try:
             result = self.imageItemModel.getAssociatedImage(item, image, **params)
-        except TileGeneralException as e:
+        except TileGeneralError as e:
             raise RestException(e.args[0], code=400)
         if not isinstance(result, tuple):
             return result
@@ -1068,3 +1088,253 @@ class TilesItemResource(ItemResource):
         if pilImage.info:
             result['info'] = pilImage.info
         return result
+
+    _tileFramesParams = [
+        ('framesAcross', int),
+        ('frameList', str),
+        ('left', float, 'region', 'left'),
+        ('top', float, 'region', 'top'),
+        ('right', float, 'region', 'right'),
+        ('bottom', float, 'region', 'bottom'),
+        ('regionWidth', float, 'region', 'width'),
+        ('regionHeight', float, 'region', 'height'),
+        ('units', str, 'region', 'units'),
+        ('unitsWH', str, 'region', 'unitsWH'),
+        ('width', int, 'output', 'maxWidth'),
+        ('height', int, 'output', 'maxHeight'),
+        ('fill', str),
+        ('magnification', float, 'scale', 'magnification'),
+        ('mm_x', float, 'scale', 'mm_x'),
+        ('mm_y', float, 'scale', 'mm_y'),
+        ('exact', bool, 'scale', 'exact'),
+        ('frame', int),
+        ('encoding', str),
+        ('jpegQuality', int),
+        ('jpegSubsampling', int),
+        ('tiffCompression', str),
+        ('style', str),
+        ('resample', 'boolOrInt'),
+        ('contentDisposition', str),
+        ('contentDispositionFileName', str)
+    ]
+
+    @describeRoute(
+        Description('Composite thumbnails of multiple frames into a single image.')
+        .param('itemId', 'The ID of the item.', paramType='path')
+        .param('framesAcross', 'How many frames across', required=False, dataType='int')
+        .param('frameList', 'Comma-separated list of frames', required=False)
+        .param('cache', 'Cache the results for future use', required=False,
+               dataType='boolean', default=False)
+        .param('left', 'The left column (0-based) of the region to process.  '
+               'Negative values are offsets from the right edge.',
+               required=False, dataType='float')
+        .param('top', 'The top row (0-based) of the region to process.  '
+               'Negative values are offsets from the bottom edge.',
+               required=False, dataType='float')
+        .param('right', 'The right column (0-based from the left) of the '
+               'region to process.  The region will not include this column.  '
+               'Negative values are offsets from the right edge.',
+               required=False, dataType='float')
+        .param('bottom', 'The bottom row (0-based from the top) of the region '
+               'to process.  The region will not include this row.  Negative '
+               'values are offsets from the bottom edge.',
+               required=False, dataType='float')
+        .param('regionWidth', 'The width of the region to process.',
+               required=False, dataType='float')
+        .param('regionHeight', 'The height of the region to process.',
+               required=False, dataType='float')
+        .param('units', 'Units used for left, top, right, bottom, '
+               'regionWidth, and regionHeight.  base_pixels are pixels at the '
+               'maximum resolution, pixels and mm are at the specified '
+               'magnfication, fraction is a scale of [0-1].', required=False,
+               enum=sorted(set(TileInputUnits.values())),
+               default='base_pixels')
+
+        .param('width', 'The maximum width of the output image in pixels.',
+               required=False, dataType='int')
+        .param('height', 'The maximum height of the output image in pixels.',
+               required=False, dataType='int')
+        .param('fill', 'A fill color.  If output dimensions are specified and '
+               'fill is specified and not "none", the output image is padded '
+               'on either the sides or the top and bottom to the requested '
+               'output size.  Most css colors are accepted.', required=False)
+        .param('magnification', 'Magnification of the output image.  If '
+               'neither width for height is specified, the magnification, '
+               'mm_x, and mm_y parameters are used to select the output size.',
+               required=False, dataType='float')
+        .param('mm_x', 'The size of the output pixels in millimeters',
+               required=False, dataType='float')
+        .param('mm_y', 'The size of the output pixels in millimeters',
+               required=False, dataType='float')
+        .param('exact', 'If magnification, mm_x, or mm_y are specified, they '
+               'must match an existing level of the image exactly.',
+               required=False, dataType='boolean', default=False)
+        .param('frame', 'For multiframe images, the 0-based frame number.  '
+               'This is ignored on non-multiframe images.', required=False,
+               dataType='int')
+        .param('encoding', 'Output image encoding.  TILED generates a tiled '
+               'tiff without the upper limit on image size the other options '
+               'have.  For geospatial sources, TILED will also have '
+               'appropriate tagging.', required=False,
+               enum=['JPEG', 'PNG', 'TIFF', 'TILED'], default='JPEG')
+        .param('jpegQuality', 'Quality used for generating JPEG images',
+               required=False, dataType='int', default=95)
+        .param('jpegSubsampling', 'Chroma subsampling used for generating '
+               'JPEG images.  0, 1, and 2 are full, half, and quarter '
+               'resolution chroma respectively.', required=False,
+               enum=['0', '1', '2'], dataType='int', default='0')
+        .param('tiffCompression', 'Compression method when storing a TIFF '
+               'image', required=False,
+               enum=['none', 'raw', 'lzw', 'tiff_lzw', 'jpeg', 'deflate',
+                     'tiff_adobe_deflate'])
+        .param('style', 'JSON-encoded style string', required=False)
+        .param('resample', 'If false, an existing level of the image is used '
+               'for the region.  If true, the internal values are '
+               'interpolated to match the specified size as needed.  0-3 for '
+               'a specific interpolation method (0-nearest, 1-lanczos, '
+               '2-bilinear, 3-bicubic)', required=False,
+               enum=['false', 'true', '0', '1', '2', '3'])
+        .param('contentDisposition', 'Specify the Content-Disposition response '
+               'header disposition-type value.', required=False,
+               enum=['inline', 'attachment'])
+        .param('contentDispositionFilename', 'Specify the filename used in '
+               'the Content-Disposition response header.', required=False)
+        .produces(ImageMimeTypes)
+        .errorResponse('ID was invalid.')
+        .errorResponse('Read access was denied for the item.', 403)
+        .errorResponse('Insufficient memory.')
+    )
+    @access.public(cookie=True)
+    @loadmodel(model='item', map={'itemId': 'item'}, level=AccessType.READ)
+    def tileFrames(self, item, params):
+        cache = params.pop('cache', False)
+        checkAndCreate = False if cache else 'nosave'
+        _adjustParams(params)
+
+        params = self._parseParams(params, True, self._tileFramesParams)
+        _handleETag('tileFrames', item, params)
+        if 'frameList' in params:
+            params['frameList'] = [
+                int(f.strip()) for f in str(params['frameList']).lstrip(
+                    '[').rstrip(']').split(',')]
+        setResponseTimeLimit(86400)
+        try:
+            result = self.imageItemModel.tileFrames(
+                item, checkAndCreate=checkAndCreate, **params)
+        except TileGeneralError as e:
+            raise RestException(e.args[0])
+        except ValueError as e:
+            raise RestException('Value Error: %s' % e.args[0])
+        if not isinstance(result, tuple):
+            return result
+        regionData, regionMime = result
+        self._setContentDisposition(
+            item, params.get('contentDisposition'), regionMime, 'tileframes',
+            params.get('contentDispositionFilename'))
+        setResponseHeader('Content-Type', regionMime)
+        if isinstance(regionData, pathlib.Path):
+            BUF_SIZE = 65536
+
+            def stream():
+                try:
+                    with regionData.open('rb') as f:
+                        while True:
+                            data = f.read(BUF_SIZE)
+                            if not data:
+                                break
+                            yield data
+                finally:
+                    regionData.unlink()
+            return stream
+        setRawResponse()
+        return regionData
+
+    @describeRoute(
+        Description('Get parameters for using tile_frames as background sprite images.')
+        .param('itemId', 'The ID of the item.', paramType='path')
+        .param('format', 'Optional format parameters, such as "encoding=JPEG&'
+               'jpegQuality=85&jpegSubsampling=1".  If specified, these '
+               'replace the defaults.', required=False)
+        .param('query', 'Addition query parameters that would be passed to '
+               'tile endpoints, such as style.', required=False)
+        .param('frameBase', 'Starting frame number (default 0).  If c/z/t/xy '
+               'then step through values from 0 to number of that axis - 1.  '
+               'The axis specification in only useful for cache reporting or '
+               'scheduling',
+               required=False, dataType='int')
+        .param('frameStride', 'Only use every frameStride frame of the image '
+               '(default 1).  c/z/t/xy to use the length of that axis',
+               required=False, dataType='int')
+        .param('frameGroup', 'Group frames when using multiple textures to '
+               'keep boundaries at a multiple of the group size number.  '
+               'c/z/t/xy to use the length of that axis.',
+               required=False, dataType='int')
+        .param('frameGroupFactor', 'Ignore grouping if the resultant images '
+               'would be more than this factor smaller than without grouping '
+               '(default 4)', required=False, dataType='int')
+        .param('frameGroupStride', 'Reorder frames based on the to stride '
+               '(default 1).  "auto" to use frameGroup / frameStride if that '
+               'value is an integer.',
+               required=False, dataType='int')
+        .param('maxTextureSize', 'Maximum texture size in either dimension.  '
+               'This should be the smaller of a desired value and of the '
+               'intended graphics environment texture buffer (default 16384).',
+               required=False, dataType='int')
+        .param('maxTextures', 'Maximum number of textures to use (default 1).',
+               required=False, dataType='int')
+        .param('maxTotalTexturePixels', 'Limit the total area of all combined '
+               'textures (default 2**30).',
+               required=False, dataType='int')
+        .param('alignment', 'Individual frame alignment within a texture.  '
+               'Used to avoid jpeg artifacts from crossing frames (default '
+               '16).',
+               required=False, dataType='int')
+        .param('maxFrameSize', 'If specified, frames will never be larger '
+               'than this, even if the texture size allows it (default None).',
+               required=False, dataType='int')
+        .param('cache', 'Report on or request caching the resultant frames.  '
+               'Scheduling creates a local job.',
+               required=False,
+               enum=['none', 'report', 'schedule'])
+        .errorResponse('ID was invalid.')
+        .errorResponse('Read access was denied for the item.', 403)
+    )
+    @access.public(cookie=True)
+    @loadmodel(model='item', map={'itemId': 'item'}, level=AccessType.READ)
+    def tileFramesQuadInfo(self, item, params):
+        metadata = self.imageItemModel.getMetadata(item)
+        options = self._parseParams(params, False, [
+            ('format', str),
+            ('query', str),
+            ('frameBase', str),
+            ('frameStride', str),
+            ('frameGroup', str),
+            ('frameGroupFactor', int),
+            ('frameGroupStride', str),
+            ('maxTextureSize', int),
+            ('maxTextures', int),
+            ('maxTotalTexturePixels', int),
+            ('alignment', int),
+            ('maxFrameSize', int),
+        ])
+        for key in {'format', 'query'}:
+            if key in options:
+                options[key] = dict(urllib.parse.parse_qsl(options[key]))
+        result = large_image.tilesource.utilities.getTileFramesQuadInfo(metadata, options)
+        if params.get('cache') in {'report', 'schedule'}:
+            needed = []
+            result['cached'] = []
+            for src in result['src']:
+                tfParams = self._parseParams(src, False, self._tileFramesParams)
+                if 'frameList' in tfParams:
+                    tfParams['frameList'] = [
+                        int(f.strip()) for f in str(tfParams['frameList']).lstrip(
+                            '[').rstrip(']').split(',')]
+                result['cached'].append(self.imageItemModel.tileFrames(
+                    item, checkAndCreate='check', **tfParams))
+                if not result['cached'][-1]:
+                    needed.append(tfParams)
+            if params.get('cache') == 'schedule' and not all(result['cached']):
+                result['scheduledJob'] = str(self.imageItemModel._scheduleTileFrames(
+                    item, needed, self.getCurrentUser())['_id'])
+        return result
